mod serial;

#[cfg(test)]
mod tests;

use chrono::Utc;
use derive_more::From;
use poise::serenity_prelude::{self, UserId, Http};
use regex::Regex;
use reqwest::StatusCode;
use rocket::http::{Cookie, CookieJar};
use serial::*;

use serde::{Deserialize, Serialize};

use crate::cookies::{token::*, user::*};

pub trait Username {
    fn username(&self) -> String;
}

#[derive(Default, Deserialize, Debug)]
pub struct User {
    #[serde(deserialize_with = "deserialize_id")]
    pub id: u64,
    #[serde(rename = "username")]
    pub name: String,
    #[serde(deserialize_with = "deserialize_discriminator")]
    pub discriminator: u16,
    pub avatar: Option<String>,
    pub deleted: bool,
}

impl Username for User {
    fn username(&self) -> String {
        if self.discriminator == 0 {
            return self.name.clone();
        }
        format!("{}#{:0>4}", self.name, self.discriminator)
    }
}

fn is_name_deleted(name: &str) -> bool {
    Regex::new(r"Deleted User [a-f0-9]{8}").unwrap().is_match(name)
}

impl User {
    pub fn from_username(username: &str) -> Self {
        let (name, discriminator) = {
            let mut iter = username.split('#');
            let name = iter.next().unwrap().to_owned();
            let discriminator = iter
                .next()
                .map(|str| str.parse().unwrap())
                .unwrap_or(0);
            (name, discriminator)
        };
        Self {
            id: 0,
            name,
            discriminator,
            avatar: None,
            deleted: false,
        }
    }

    pub async fn fetch(http: &Http, id: u64) -> serenity_prelude::Result<Self> {
        let user = UserId(id).to_user(http).await?;
        Ok(Self {
            id,
            avatar: user.avatar,
            deleted: is_name_deleted(&user.name),
            name: user.name,
            discriminator: user.discriminator,
        })
    }

    fn avatar(&self) -> String {
        match &self.avatar {
            Some(avatar) => format!(
                "https://cdn.discordapp.com/avatars/{}/{}.{}?size=1024",
                self.id,
                avatar,
                if avatar.starts_with("a_") { "gif" } else { "webp" }
            ),
            // Archived user or user with no avatar, calculate default avatar
            // https://www.reddit.com/r/discordapp/comments/au6v4e/comment/eh61dm6/
            // https://docs.rs/serenity/0.11.5/serenity/model/user/struct.User.html#method.default_avatar_url
            None => format!("https://cdn.discordapp.com/embed/avatars/{}.png", self.discriminator % 5),
        }
    }
}

#[derive(Default, Serialize)]
pub struct SessionUser(pub User);

impl Username for SessionUser {
    fn username(&self) -> String {
        self.0.username()
    }
}

impl SessionUser {
    pub async fn init(token: &str, cookies: &CookieJar<'_>) -> Result<Self, GetUserError> {
        let (status, text) = {
            let response = reqwest::Client::new()
                .get("https://discord.com/api/users/@me")
                .header("Authorization", format!("Bearer {token}"))
                .send()
                .await?;
            (response.status(), response.text().await)
        };
        if !status.is_success() {
            return Err(GetUserError::Discord {
                status,
                message: text.ok(),
            });
        }
        let user: User = serde_json::from_str(&text?)?;
        cookies.add_private(Cookie::new(TOKEN_COOKIE, token.to_owned()));
        cookies.add_private(Cookie::new(USER_ID_COOKIE, user.id.to_string()));
        cookies.add_private(Cookie::new(USER_NAME_COOKIE, user.name.clone()));
        cookies.add_private(Cookie::new(
            USER_DISCRIMINATOR_COOKIE,
            user.discriminator.to_string(),
        ));
        cookies.add_private(Cookie::new(USER_AVATAR_COOKIE, user.avatar()));
        Ok(Self(user))
    }

    pub fn purge(cookies: &CookieJar<'_>) {
        cookies.remove_private(Cookie::named(TOKEN_COOKIE));
        cookies.remove_private(Cookie::named(USER_ID_COOKIE));
        cookies.remove_private(Cookie::named(USER_NAME_COOKIE));
        cookies.remove_private(Cookie::named(USER_DISCRIMINATOR_COOKIE));
        cookies.remove_private(Cookie::named(USER_AVATAR_COOKIE));
        cookies.remove(Cookie::named(TOKEN_EXPIRE_COOKIE));
    }

    fn from_cookies(cookies: &CookieJar<'_>) -> Option<Self> {
        Some(Self(User {
            id: parse_cookie_value(cookies, USER_ID_COOKIE)?,
            name: parse_cookie_value(cookies, USER_NAME_COOKIE)?,
            discriminator: parse_cookie_value(cookies, USER_DISCRIMINATOR_COOKIE)?,
            avatar: Some(parse_cookie_value(cookies, USER_AVATAR_COOKIE)?),
            deleted: false,
        }))
    }

    pub async fn get(cookies: &CookieJar<'_>) -> Result<Option<Self>, GetUserError> {
        let user = match Self::from_cookies(cookies) {
            Some(user) => user,
            None => return Ok(None),
        };
        if cookies
            .get(TOKEN_EXPIRE_COOKIE)
            .map(|expire| expire.value().parse::<i64>())
            .and_then(Result::ok)
            .map_or(true, |timestamp| Utc::now().timestamp() >= timestamp)
        {
            cookies.remove_private(Cookie::named(TOKEN_COOKIE));
            cookies.remove_private(Cookie::named(USER_ID_COOKIE));
            cookies.remove_private(Cookie::named(USER_NAME_COOKIE));
            cookies.remove_private(Cookie::named(USER_DISCRIMINATOR_COOKIE));
            cookies.remove_private(Cookie::named(USER_AVATAR_COOKIE));
            cookies.remove(Cookie::named(TOKEN_EXPIRE_COOKIE));
            return Ok(None);
        }
        Ok(Some(user))
    }
}

#[derive(From, Debug)]
pub enum GetUserError {
    Reqwest(reqwest::Error),
    Deserialize(serde_json::Error),
    #[allow(unused)]
    Discord {
        status: StatusCode,
        message: Option<String>,
    },
}

fn parse_cookie_value<T: std::str::FromStr>(cookies: &CookieJar<'_>, name: &str) -> Option<T> {
    cookies.get_private(name)?.value().parse().ok()
}
